/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#include "flag.h"
#include "oi_wavelength.h"
#include "oi_vis.h"
#define clight  2.9979250e8 /*m per sec */
#define twopi   6.2831853072

/* external int errno;*/
/*
    Usage::
    oirRedCal Tag 

    Output: a FitsFile Tag.redcal.fits

    containing: 
    1. a primary header (copied from Tag.corr.fits)
    2. a wavelength table (copied from Tag.corr.fits)
    3. a oi_vis table with 1 row: 
       instrumental visibility (corr flux/masked photom flux)
*/


int main (int argc, char** argv) {

   int i, j, ierr;
   long nRow, iRow ;
   int nx, nWave ;

   fitsfile *inFile;
   fitsfile *outFile = NULL;
   int iKey ;

   imaging_fdata_table_def *idata;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_row *inData;

   wavelength_table_def *wTable;
   wavelength_table_row *wRow;
   float *wave, *calFlux;

   flag_table_def *fTable;
   flag_table_row *fRow;

   float *photoMask    ; /* Photometry under mask */
   float *photoMaskErr ; /* mask photometry error estimate */
   float *visAmp;        /* amplitude Cal Vis */
   float *visPhase;      /* phase average Vis(deg) */
   float *visAmpErr;     /* amplitude Cal Vis rms errors*/
   float *visPhaseErr;   /* phase rms (deg) errors */
   float *insVis;        /* instrumental Visibility */
   float *insVisErr;     /* instrumental Visibility Err */
   float maxPhot;

   vis_table_def *vTable;
   vis_table_row *vRow;

   char *photo= ".photometry.fits";
   char *corr  = ".corr.fits";
   char *cal  = ".redcal.fits";

   char inFileName[130];
   char outFileName[130];
   float p, x, y;

   char *bang = "!";

   ierr = 0;
/*
      open calibrator photometry file 
*/
   strcpy(inFileName, argv[1]);
   strcat(inFileName, photo);

   fits_open_file (&inFile, inFileName, READONLY, &ierr);
   if (ierr != 0) {
      printf("fits open cal phot on %s gave error %i\n", inFileName,
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
/*
      get detector info
*/

   DetectorDef = read_imaging_detector_table(inFile, &ierr);
   idata          = open_imaging_fdata_table(inFile, &ierr);
   inData         = make_imaging_fdata_row(idata);

/*
      allocate float arrays
*/
   nx = DetectorDef->naxis[0][0];
   photoMask      = (float *)malloc(nx * sizeof(float));
   photoMaskErr   = (float *)malloc(nx * sizeof(float));
   visAmp         = (float *)malloc(nx * sizeof(float));
   visPhase       = (float *)malloc(nx * sizeof(float));
   visAmpErr      = (float *)malloc(nx * sizeof(float));
   visPhaseErr    = (float *)malloc(nx * sizeof(float));
   wave           = (float *)malloc(nx * sizeof(float));
   calFlux        = (float *)malloc(nx * sizeof(float));
   insVis         = (float *)malloc(nx * sizeof(float));
   insVisErr      = (float *)malloc(nx * sizeof(float));

/*
     read 6th row (AB total/masked photometry)+
          12th     (rms errors)
*/
   i = read_imaging_fdata_table(inFile, idata, inData, 6, &ierr);
   memcpy(photoMask, inData->data[0], nx*sizeof(float));
   i = read_imaging_fdata_table(inFile, idata, inData, 12, &ierr);
   memcpy(photoMaskErr, inData->data[0], nx*sizeof(float));
   i = close_imaging_fdata_table(inFile, idata, &ierr);
   kill_imaging_fdata_row(inData);

/* 
   open calibrator correlated flux table 
   create output file and copy header and wavelength
   info, then and (scalar) average all rows 
*/
   strcpy(inFileName, argv[1]);
   strcat(inFileName, corr);

   fits_open_file (&inFile, inFileName, READONLY, &ierr);
   if (ierr != 0) {
      printf("fits open cal flux on %s gave error %i\n", inFileName,
         ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }

   strcpy(outFileName, bang);
   strcat(outFileName, argv[1]);
   strcat(outFileName, cal);

/*
      create output file and copy header
*/
   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n",argv[1], ierr);
      return 1;
   }
   i = fits_copy_header (inFile, outFile, &ierr);



/*
     get wavelength info
*/
   wTable = open_wavelength_table(inFile, &ierr);
/*
     copy  wavelength table to output file
*/
   i = fits_copy_hdu(inFile, outFile, 0, &ierr);
   wRow   = make_wavelength_table_row(wTable);
   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
   if (nRow != nx) {
      printf("Number of rows in wavelength table %li != nx=%i\n",
         nRow, nx);
      return 1;
   }
   for (iRow =1;iRow < nRow; iRow++) {
      i = read_wavelength_table(inFile, wTable, wRow, iRow, &ierr);
      wave[iRow-1] = wRow->wave;
      }
   kill_wavelength_table_def(wTable);
   kill_wavelength_table_row(wRow);

/*
      open input vis table section
*/
   vTable = open_vis_table(inFile, &ierr);
   if (ierr != 0) {
      printf("fits open cal flux table error %i\n", ierr);
      return 1;
   }
   vRow = make_vis_row(vTable);
/*
     number of entries in cal vis table
*/
   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
   memset(visAmp,   0, nx*sizeof(float));
   memset(visPhase, 0, nx*sizeof(float));
   memset(visAmpErr,   0, nx*sizeof(float));
   memset(visPhaseErr, 0, nx*sizeof(float));
   for (iRow=1; iRow<=nRow; iRow++){
      i = read_vis_table_row(inFile, vTable, vRow, iRow, &ierr);
      for (i=0; i<nx; i++) {
         visAmp[i]   = visAmp[i]+ vRow->visAmp[i];
         visPhase[i] = visPhase[i]+ vRow->visPhi[i];
         visAmpErr[i]   = visAmpErr[i]+ vRow->visAmpErr[i]*vRow->visAmpErr[i];
         visPhaseErr[i] = visPhaseErr[i]+ vRow->visPhiErr[i]*vRow->visPhiErr[i];
      }
   }
   for (i=0; i<nx; i++) {
      visAmp[i]   = visAmp[i]/nRow;
      visPhase[i] = visPhase[i]/nRow;
      visAmpErr[i]   = sqrt(visAmpErr[i])/nRow;
      visPhaseErr[i] = sqrt(visPhaseErr[i])/nRow;
   }
/*
      close input vis table, also closes file
*/
   i = close_vis_table(inFile, vTable, &ierr);
   kill_vis_row(vRow);

/*
      calculate instrumental visibility = corrflux/photometry; 
      first find maximum of photometry and take .5% of
      this as maximum plausible photometry (to avoid 0 divides)

      also calculate photometry calibration = flux/counts
      and store in calFlux;
*/
   maxPhot = -1.e9;
   for (i=0;i<nx;i++) if (photoMask[i] > maxPhot) maxPhot = photoMask[i];
   maxPhot = .005 * maxPhot;
   for (i=0;i<nx;i++) {
      if (photoMask[i] > maxPhot) p = photoMask[i]; else p = maxPhot;
      insVis[i]     = visAmp[i]/p;
      x = visAmpErr[i];
      y = photoMaskErr[i];
      insVisErr[i]  = sqrt(x*x+y*y*insVis[i]*insVis[i])/p;
   }

/*
      create output table
      1st row -> ins visibility (and phase)
      2nd row -> spcal (phase = 0)
      3rd row -> "strehl ratio (phase = 0);
*/
   vTable = make_vis_table_def(outFile, nx, "1776-07-04", "VLTI",
      "MIDI/PRISM", &ierr);
   vRow = make_vis_row(vTable);
   /*memcpy(vRow->visAmp, visAmp, nx*sizeof(float));*/
   memcpy(vRow->visAmp,    insVis, nx*sizeof(float));
   memcpy(vRow->visAmpErr, insVisErr, nx*sizeof(float));
   memcpy(vRow->visPhi,    visPhase, nx*sizeof(float));
   memcpy(vRow->visPhiErr, visPhaseErr, nx*sizeof(float));
   i = write_vis_table_row(outFile, vTable, vRow, 1, &ierr);

/*
      shut down and shut up
*/
   i = close_vis_table(outFile, vTable, &ierr);
   kill_vis_row(vRow);

   free(photoMask);
   free(photoMaskErr);
   free(visAmp);
   free(visPhase);
   free(visAmpErr);
   free(visPhaseErr);
   free(wave);
   free(calFlux);
   free(insVis);
   free(insVisErr);
   return 0;
}
