/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#include "delay.h"
#include "dispersion.h"
#define clight  2.9979250e8 /*m per sec */
#define twopi   6.2831853072
#define radeg  57.29578
/*
*/
#define DOPHASE
#define DOGROUP
#define DOOPD
/*
#define SMOOTHSIGMA 6.
*/
#define SMOOTHSIGMA 1.5
/*
    Usage::
    oirRotateGroupDelay infile delayfile outfile 
*/
void Usage() {
   printf("   oirRotateGroupDelay infile delayfile outfile [-phase phasefile] [-smooth smooth]\n");
   printf("      infile = file name of output of oirFormFringes\n");
   printf("      delayfile = file name of output of oirGroupDelay\n");
   printf("      outfile = name of output file\n");
   printf("      phasefile = name of file containing previous solved phase\n");
   printf("         optional; if not found phases will be found from data\n");
   printf("      smooth (optional) = size of smoothing window before determing\n");
   printf("      water vapor dispersion phase\n");
}

/*
 *            sum rows in storage buffer with weights given
 *            by gaussians wrt row number centered on oRow.
 *            Put result in array smooth.
 *            Return the total of all weights
 */
float smoothData(imaging_fdata_row **rows, 
   int *inRows, int oRow, long npix, int nFrame, float sigma, float *smooth) {

   int i;
   long ipix, nx2;
   int nf2 ;
   float weight, totalWeight ;
   float x;

   nf2 = 2*nFrame + 1;
   nx2 = 2*npix;

   memset(smooth, 0, nx2 * sizeof(float)); /* zero accumulator */

   totalWeight = 0.;
   for (i=0; i<nf2; i++) if (inRows[i] > 0){
      x = (inRows[i]-oRow)/sigma;
      weight = exp(-0.5* x*x);
      totalWeight = totalWeight + weight;
      for (ipix=0;ipix<nx2;ipix++) smooth[ipix] = 
         smooth[ipix] + weight * rows[i]->data[0][ipix];
   }
   return totalWeight;
} /* end of smoothData */

void rotateDataPhase(float *data, int nx, int ny, float phase) {
   float c,s,w;
   int i,j;

   c = cos(phase);
   s = sin(phase);
   for (j=0;j<ny;j++) for (i=0;i<nx; i++) {
      w                  = c*data[2*i+2*j*nx]   - s*data[2*i+2*j*nx+1];
      data[2*i+2*j*nx+1] = c*data[2*i+2*j*nx+1] + s*data[2*i+2*j*nx];
      data[2*i+2*j*nx]   = w;
   }
   return;
} /* end of rotateData Phase */

void removeDataPhase(float *smData, float *data, 
   int nx, int ny, float *phase, int oRow){
   float c,s,w;
   float cTotal, sTotal;
   int y1, y2;
   int i,j;

   cTotal = 0;
   sTotal = 0;
   if (ny ==1) {
      y1 = 0;
      y2 = 0;
   } else {
      y1 = 10;
      y2 = 30;
      if (y1 <1)  y1 = 1;
      if (y2 >ny) y2 = ny-1;
   }
/*
 *        vTotal = (cTotal, sTotal) is complex sum over wavelength 
 *         of all the pixels in smoothed data row
 */
   for (j=y1; j<=y2; j++) for (i=0; i<nx; i++) {
      cTotal = cTotal + smData[2*i+2*j*nx];
      sTotal = sTotal + smData[2*i+2*j*nx+1];
   }
   phase[oRow-1] = radeg * atan2(-sTotal,cTotal); /* write phase into table */
/*
 *        we will calculate phase to be removed from this row as
 *        phi = Atan(sTotal, cTotal) !!BUT!! this introduces an artifical
 *        tendency to drive measured phases -> 0 even if they are pure
 *        noise, because we used data from a given point to correct
 *        phase from the same point.  We attempt to remove this effect
 *        by, for each pixel, removing the contribution of this pixel
 *        to vTotal before calculating phi for this pixel
 */
/*
 *        loop over pixels
 */
   for (j=0; j<ny; j++) for (i=0; i<nx; i++) {
/*             remove influence of this pixel */

      if ((j >=y1) && (j <= y2)) {
         s= sTotal - smData[2*i+2*j*nx+1];
         c= cTotal - smData[2*i+2*j*nx];
      }

      w = sqrt(c*c + s*s);  /* absolute value */
      c =  c/w;             /* sine and cosine */
      s = -s/w;
/*                             now remove phase from data */
      w                  = c*data[2*i+2*j*nx]   - s*data[2*i+2*j*nx+1];
      data[2*i+2*j*nx+1] = c*data[2*i+2*j*nx+1] + s*data[2*i+2*j*nx];
      data[2*i+2*j*nx]   = w;
   }

   return;
} /* end of removeDataPhase */

int main (int argc, char** argv) {

   int i, j, ierr;
   int iy, y1, y2;
   long nRow, iRow, oRow, pRow;
   int nx, ny;
   long npix, ii;
   float *frequency, *waveNo;
   double opd;


   char outFileName[130];
   char *bang="!";

   fitsfile *inFile;
   fitsfile *outFile ;
   fitsfile *phaseFile ;
   int iKey, maxstep;

   imaging_fdata_table_def *idata;
   imaging_fdata_table_def *odata;
   imaging_detector_table_def *outDetectorDef;
   imaging_fdata_row *inData;
   imaging_fdata_row **outDataArray, *outData;

   delay_table_row *delayData;
   delay_table_def *dTable;
   double* groupDelay;
   double* delayTime;
  
   
   float *phase, c,s,w;

   float smBufferWidth, smBW2;
   float *outBuffer, *saveBuffer, *od;
   float smoothWeight;
   float sigma;
   int   *smBufferI, smI;

   dispersion_table_def *disTable;
   dispersion_table_row *disRow;

   char* phaseFileName = NULL;
   outFile             = NULL;
   phaseFile           = NULL;
   int   iarg      = 4;
   ierr = 0;
   sigma = SMOOTHSIGMA;

   if (argc < 3) {
      Usage();
      return 1;
   }
   while (iarg < argc ) {
      if (!strcmp("-phase", argv[iarg])) phaseFileName = argv[++iarg];
      if (!strcmp("-smooth", argv[iarg])) sscanf(argv[++iarg],"%f", &sigma);
      ++iarg;
   }
   printf("      ******   oirRotateGroupDelay   ******\n");
   printf("   input fringe file: %s\n", argv[1]);
   printf("   output fringe file: %s\n", argv[3]);
   printf("   group delay information from: %s\n", argv[2]);
   if (phaseFileName == NULL) 
      printf("   phase information derived from data.\n\n");
      else printf("   phase information from: %s\n\n", phaseFileName);
   printf("   smoothing sigma is %f\n",sigma);

/*
      set up smoothing buffers for dispersion removal
*/
   smBW2         = (int)(.5+sigma);
   smBufferWidth = 1 + 2* smBW2;
   smBufferI     = (int *)malloc(smBufferWidth *sizeof(int));
   memset(smBufferI, 0, smBufferWidth*sizeof(int));
   outDataArray = (imaging_fdata_row **) 
      malloc (smBufferWidth * sizeof(imaging_fdata_row *));
/*
      open delay file and get delay data
*/
   fits_open_file (&inFile, argv[2], READONLY, &ierr);
   if (ierr != 0) {
      printf("error %i opening delay file %s\n", ierr, argv[2]);
      printf("%s \n",strerror(errno));
      return 1;
   }
   dTable = open_delay_table(inFile, &ierr);
   if (ierr != 0) {
      printf("error %i opening delay table \n", ierr);
      return 1;
   }
   delayData = make_delay_table_row(dTable);
   i= fits_get_num_rows(inFile, &nRow, &ierr);
   groupDelay = (double *)malloc(nRow*sizeof(double));;
   delayTime  = (double *)malloc(nRow*sizeof(double));;
   for (iRow =1; iRow<=nRow;iRow++) {
      i=read_delay_table(inFile, dTable, delayData, iRow, &ierr);
      if (ierr != 0) {
         printf("error %i reading delay table row %li\n", ierr, iRow);
         return 1;
      }
      groupDelay[iRow-1] = delayData->delay*clight; /*meters*/
      delayTime[iRow-1]  = delayData->time; /*meters*/
   }
/*
      close file and clean up structures
*/
   i = close_delay_table(inFile, dTable, &ierr);
   kill_delay_table_row(delayData);
/*
      open input file 
*/
   fits_open_file (&inFile, argv[1], READONLY, &ierr);
   if (ierr != 0) {
      printf("error %i opening data file %s\n", ierr, argv[1]);
      printf("%s \n",strerror(errno));
      return 1;
   }
/*
      create output data table 
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[3]);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n",argv[3], ierr);
      printf("%s \n",strerror(errno));
      return 1;
      }

/*
      copy header from input->output
*/
   i = fits_copy_header(inFile, outFile, &ierr);

/*
      get detector information
*/
   outDetectorDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr != 0) {
      printf("error %i reading detector table \n", ierr);
      return 1;
   }
   idata          = open_imaging_fdata_table(inFile, &ierr);
   if (ierr != 0) {
      printf("error %i reading data table \n", ierr);
      return 1;
   }
   iKey = table_find_key(idata->def, "MAXSTEP");
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&maxstep,
      &ierr);

/*     
   get row size, then double it (pseudo-complex)
*/
   nx = outDetectorDef->naxis[0][0];
   ny = outDetectorDef->naxis[0][1];
   npix = (long)nx * ny;
   outDetectorDef->naxis[0][0] = 2*nx; 

/*
      create output table definitions
*/

   odata = make_imaging_fdata_table_def (outFile, 
      outDetectorDef, maxstep, &ierr); 
   if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return 1;
      }
   inData = make_imaging_fdata_row(idata);
   for (i=0;i<smBufferWidth;i++) outDataArray[i] = make_imaging_fdata_row(odata);
   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
/*       
      get frequency information from detector table
*/
   frequency = imaging_detector_wcs(outDetectorDef, 1);
   waveNo = (float *)malloc(nx * sizeof(float));
   for (i=0; i<nx; i++) waveNo[i] = twopi * frequency[i]/clight;
   phase = (float *)malloc(nRow * sizeof(float));
/*
 *        if phase specified externally, go get it
*/
   if (phaseFileName != NULL) {
      fits_open_file (&phaseFile, phaseFileName, READONLY, &ierr);
      if (ierr != 0) {
         printf("error %i opening phase file %s\n", ierr, phaseFileName);
         printf("%s \n",strerror(errno));
         return 1;
      }
      disTable = open_dispersion_table(phaseFile, &ierr);
      if (ierr != 0) {
         printf("error %i opening phase table in %s\n", ierr, phaseFileName);
         printf("%s \n",strerror(errno));
         return 1;
      }
      disRow = make_dispersion_table_row(disTable);
      for (iRow=1; iRow<=nRow; iRow++) {
         i = read_dispersion_table(phaseFile, disTable, disRow,
            iRow, &ierr);
         phase[iRow-1] = -disRow->dispersion/radeg; 
      }
      kill_dispersion_table_def(disTable);
      kill_dispersion_table_row(disRow);
      i = fits_close_file(phaseFile, &ierr);
   }
/*       
         loop over input data, put in cyclic smoothing buffer,
         multiplying input value * exp(i*k*(D-opd0))
*/
   smI  = -1;
   oRow = 0;
   outBuffer  = (float *)malloc(2 * npix *sizeof(float));
   saveBuffer = (float *)malloc(2 * npix *sizeof(float));

   for (iRow=1; iRow<=nRow; iRow++) {
/*
        read new row, multiply by cos/sin of phase and store
*/
      i = read_imaging_fdata_table(inFile, idata, inData, iRow, &ierr);
      if (ierr != 0) {
         printf("error %i reading data table row %li\n", ierr, iRow);
         return 1;
      }
/*
            is data OK?
*/
      if (!isnan(inData->time)) {

/*
            choose output buffer and copy nonData info
*/
      if (++smI >= smBufferWidth) smI = 0;
      smBufferI[smI] = iRow;
      outData = outDataArray[smI];

      outData->time         = inData->time;
      outData->opd[0]       = inData->opd[0];
      outData->opd[1]       = inData->opd[1];
      outData->local_opd[1] = inData->local_opd[1];
      outData->local_opd[0] = inData->local_opd[0];
      outData->exp          = inData->exp;
      outData->offset[0]    = inData->offset[0];
      outData->offset[1]    = inData->offset[1];
      outData->rotation     = inData->rotation;
      outData->frame        = inData->frame;
      outData->step         = inData->step;
      outData->reference    = inData->reference;
      outData->targtype[0]  = inData->targtype[0];  
/*
         calculate opd, remove group delay, rotate fringe
*/
      opd = 0.;
#ifdef DOOPD
      opd = inData->opd[1]+inData->local_opd[1] - inData->opd[0] - inData->local_opd[0];
#endif
#ifdef DOGROUP
      opd = opd - groupDelay[iRow-1];
#endif
      for (j=0; j<nx; j++) {
         c = cos(opd * waveNo[j]);
         s = sin(opd * waveNo[j]);
         
         for (iy=0;iy<ny;iy++) {
            ii = j+iy*nx;
            outData->data[0][2*ii]   = inData->data[0][ii]* c;
            outData->data[0][2*ii+1] = inData->data[0][ii]* s;
         }
      }
/*
         which row should we output (if any)
*/
      if (iRow < smBufferWidth) pRow = (int) (.5*(iRow+1)); else pRow=iRow-smBW2;
      if (pRow > oRow ) { /* there is a new row to output */
         oRow = pRow;
/*
 *       set scratch variable pRow to be index in data array that points
 *       to row oRow in total dataset
*/
         pRow = -1;
         for (i=0; i<smBufferWidth;i++) if (smBufferI[i] == oRow) pRow = i;
/*
            copy data from pRow -> saveBuffer so we can put it back later
            (for next smoothing round);  
*/
         if (pRow >= 0) {
            outData = outDataArray[pRow];
            od      = outData->data[0];
            memcpy(saveBuffer, od, 2*npix*sizeof(float));
/*
            gaussian smooth data into outBuffer
*/
            smoothWeight = smoothData(outDataArray, smBufferI, oRow, npix, 
               smBW2, sigma, outBuffer);
/*
            if external phase rotate visibilities in od; else
            calculate from smoothed outBuffer, remove and store
*/
#ifdef DOPHASE
            if (phaseFileName != NULL) 
               rotateDataPhase(od, nx, ny, phase[oRow-1]);
            else 
               removeDataPhase(outBuffer, od, nx, ny, phase, oRow);
#endif
/*
 *          write output to file
*/
            i = write_imaging_fdata_table(outFile, odata, outData, oRow, &ierr);
            if (ierr != 0) {
               printf("error %i writing data table row %li\n", ierr, oRow);
               return 1;
            }
            if (0 == oRow % 100) fprintf(stderr, "   Frame: %d\r",oRow);

/*  
   put data back for next smoothing operation
*/
            memcpy(od, saveBuffer, 2*npix*sizeof(float));
         } /* found oRow in output buffer */
      } /* row ready to output */
   }  /* check data OK */
   }  /*  loop over input rows */

/*
        write out last few rows
*/
   for (iRow = oRow +1; iRow <=nRow; iRow++) {

      pRow = -1;
      for (i=0; i<smBufferWidth;i++) if (smBufferI[i] == iRow) pRow = i;

      if (pRow >= 0) {
        outData = outDataArray[pRow];
        od      = outData->data[0];
        memcpy(saveBuffer, od, 2*npix*sizeof(float));
        smoothWeight = smoothData(outDataArray, smBufferI, 
           iRow, npix ,smBW2, sigma, outBuffer);
        if (phaseFileName != NULL) 
           rotateDataPhase(od, nx, ny, phase[iRow-1]);
        else 
           removeDataPhase(outBuffer, od, nx, ny, phase, iRow);
        i = write_imaging_fdata_table(outFile, odata, outData, iRow, &ierr);
        if (ierr != 0) {
           printf("error %i writing data table row %li\n", ierr, iRow);
           return 1;
        }

/*  
   put data back for next smoothing operation
*/
        memcpy(od, saveBuffer, 2*npix*sizeof(float));
     } /* found oRow in output buffer */
   }  /*  loop over last output rows */
   i = close_imaging_fdata_table(inFile, idata, &ierr);
   kill_imaging_fdata_table_def(odata);
   kill_imaging_fdata_row(inData);
   for (i=0; i< smBufferWidth; i++) kill_imaging_fdata_row(outDataArray[i]);
/*
      create dispersion table if required
*/
   if (phaseFileName == NULL) {
      disTable = make_dispersion_table_def(outFile, "1776-07-04",
         "Constant Phase", &ierr);
      disRow = make_dispersion_table_row(disTable);
      for (iRow=1; iRow<=nRow; iRow++) {
         disRow->time = delayTime[iRow-1];
         disRow->telescope  = 1;
         disRow->dispersion = -phase[iRow-1];
         write_dispersion_table(outFile, disTable, disRow, iRow, &ierr);
      }
      kill_dispersion_table_def(disTable);
      kill_dispersion_table_row(disRow);
   }
   i = fits_close_file(outFile, &ierr);
   free(frequency);
   free(waveNo);
   free (groupDelay);
   free (delayTime);
   free(phase);
   free(smBufferI);
   free(outDataArray);
   free(outBuffer);
   free(saveBuffer);
   printf("\n");
   return 0;
}
