pro rdfloat,name,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13,v14,v15,v16,v17, $
            v18,v19,v20,v21,v22,v23,v24,v25,v26,v27,v28,v29,v30,v31,v32,v33, $
            v34,v35,v36,v37,v38,v39,v40,v41,v42,v43,v44,v45,v46,v47,v48,v49, $
            v50,v51,v52,v53,v54,v55,v56,v57,v58,v59,v60,v61,v62,v63,v64,v65, $
            v66,v67,v68,v69,v70,v71,v72,v73,v74,v75,v76,v77,v78,v79,v80,     $
            SKIPLINE = skipline, NUMLINE = numline,DOUBLE=double, $
            SILENT = silent, COLUMNS = columns
;+
; NAME:
;      RDFLOAT
; PURPOSE:
;      Quickly read a numeric ASCII data file into IDL floating/double vectors.  
; EXPLANATION:
;      Columns of data may be separated by tabs or spaces.      This 
;      program is fast but is restricted to data files where all columns can 
;      be read as floating point (or all double precision).   
;
;      Use READCOL if  greater flexibility is desired.   Use READFMT to read a 
;      fixed-format ASCII file.   Use FORPRINT to print columns of data.
;
; CALLING SEQUENCE:
;      RDFLOAT, name, v1, [ v2, v3, v4, v5, ...  v19] 
;                         COLUMNS, /DOUBLE, SKIPLINE = , NUMLINE = ]
;
; INPUTS:
;      NAME - Name of ASCII data file, scalar string.  In VMS, an extension of 
;              .DAT is assumed, if not supplied.
;
; OPTIONAL INPUT KEYWORDS:
;      COLUMNS - Numeric scalar or vector specifying which columns in the file
;               to read.    For example, if COLUMNS = [3,7,11] then the first
;               output variable (v1) would contain column 3, the second would
;               contain column 7 and the third would contain column 11.   If
;               the number of elements in the COLUMNS vector is less than the
;               number of output parameters, then consecutive columns are 
;               implied.    For example, if 3 output parameters are supplied
;               (v1,v2,v3) and COLUMNS = 3, then columns 3,4, and 5 will be
;               read.   
;      SKIPLINE - Integer scalar specifying number of lines to skip at the top
;              of file before reading.   Default is to start at the first line.
;      NUMLINE - Integer scalar specifying number of lines in the file to read.  
;             Default is to read the entire file
;      /DOUBLE - If this keyword is set, then all variables are read in as
;              double precision.
;      /SILENT - Set this keyword to suppress any informative messages
;
; OUTPUTS:
;      V1,V2,V3,...V19 - IDL vectors to contain columns of data.
;               Up to 19 columns may be read.  All output vectors are of type
;               float, unless the /DOUBLE keyword is set, 
;
; EXAMPLES:
;      Each row in a file 'position.dat' contains a star number and 6 columns
;      of data giving an RA and Dec in sexigesimal format.   Read into IDL 
;      variables.     
;
;       IDL> rdfloat,'position.dat',ID,hr,min,sec,deg,dmin,dsec  
;
;       All output vectors will be floating point.    To only read the 
;       declination vectors (Deg,dmin,dsec)
;
;       IDL> rdfloat,'position.dat',deg,dmin,dsec,col=4
;
; RESTRICTIONS:
;      (1) All rows in the file must be formatted identically (except for 
;          those skipped by SKIPLINE).    RDFLOAT reads the first line of 
;          the data (after SKIPLINE) to determine the number of columns of 
;          data.
;      (2) Cannot be used to read strings
; PROCEDURES USED:
;      NUMLINES()
; REVISION HISTORY:
;      Written         W. Landsman                 September 1995
;      Call NUMLINES() function                    February 1996
;      Read up to 19 columns                       August 1997
;      Converted to IDL V5.0   W. Landsman         September 1997
;      Allow to skip more than 32767 lines  W. Landsman  June 2001
;      Added /SILENT keyword   W. Landsman         March 2002
;      Added COLUMNS keyword, use STRSPLIT    W. Landsman May 2002
;      Use SKIP_LUN if V5.6 or later          W. Landsamn Nov 2002
;-     Added extra columns                    M. Nuyten Jan 2004
  On_error,2                           ;Return to caller
  FORWARD_FUNCTION strsplit            ;Pre V5.3 compatibility

  if N_params() lt 2 then begin
     print,'Syntax - RDFLOAT, name, v1, [ v2, v3,...v60 '
     print,'                    COLUMNS = ,/DOUBLE, SKIPLINE =, NUMLINE = ]'
     return
  endif

; Get number of lines in file

   nlines = NUMLINES( name )
   if nlines LT 0 then return

   if not keyword_set( SKIPLINE ) then skipline = 0
   nlines = nlines - skipline
   if keyword_set( NUMLINE) then nlines = numline < nlines

;Read first line, and determine number of columns of data

   openr, lun, name, /GET_LUN
   temp = ''
   if skipline GT 0 then $
        if !VERSION.RELEASE GE '5.6' then skip_lun, lun, skipline, /lines else $
        for i=0L,skipline-1 do readf, lun, temp
   readf,lun,temp
   
   if !VERSION.RELEASE GE '5.3' then $
   colval = strsplit(temp) else $
   colval = str_sep( strtrim( strcompress(temp),2),' ')
   ncol = N_elements(colval)

;Create big output array and read entire file into the array

   if keyword_set(DOUBLE) then bigarr = dblarr(ncol, nlines, /NOZERO) $ 
                          else bigarr = fltarr(ncol, nlines, /NOZERO) 

   close,lun
   openr, lun, name
   if skipline GT 0 then $
        if !VERSION.RELEASE GE '5.6' then skip_lun, lun, skipline, /lines else $
        for i=0L,skipline-1 do readf, lun, temp

   readf, lun, bigarr
   free_lun, lun

   if not keyword_set(SILENT) then $
       message, strtrim(nlines,2) + ' lines of data read',/INF

   Nvector = (N_params()-1) < ncol
   if N_elements(columns) EQ 0 then c = indgen(nvector) else c = columns - 1
   Nc = N_elements(c)
   if Nc LT nvector then c = [c,indgen(nvector-nc) + c[nc-1] +1 ] 
   v1 = reform( bigarr[c[0],*])
 
   if Nvector GT 1 then v2 = reform( bigarr[c[1],*]) else return
   if Nvector GT 2 then v3 = reform( bigarr[c[2],*]) else return
   if Nvector GT 3 then v4 = reform( bigarr[c[3],*]) else return
   if Nvector GT 4 then v5 = reform( bigarr[c[4],*]) else return
   if Nvector GT 5 then v6 = reform( bigarr[c[5],*]) else return
   if Nvector GT 6 then v7 = reform( bigarr[c[6],*]) else return
   if Nvector GT 7 then v8 = reform( bigarr[c[7],*]) else return
   if Nvector GT 8 then v9 = reform( bigarr[c[8],*]) else return
   if Nvector GT 9 then v10 = reform( bigarr[c[9],*]) else return
   if Nvector GT 10 then v11 = reform( bigarr[c[10],*]) else return
   if Nvector GT 11 then v12 = reform( bigarr[c[11],*]) else return
   if Nvector GT 12 then v13 = reform( bigarr[c[12],*]) else return
   if Nvector GT 13 then v14 = reform( bigarr[c[13],*]) else return
   if Nvector GT 14 then v15 = reform( bigarr[c[14],*]) else return
   if Nvector GT 15 then v16 = reform( bigarr[c[15],*]) else return
   if Nvector GT 16 then v17 = reform( bigarr[c[16],*]) else return
   if Nvector GT 17 then v18 = reform( bigarr[c[17],*]) else return
   if Nvector GT 18 then v19 = reform( bigarr[c[18],*]) else return
   if Nvector GT 19 then v20 = reform( bigarr[c[19],*]) else return
   if Nvector GT 20 then v21 = reform( bigarr[c[20],*]) else return
   if Nvector GT 21 then v22 = reform( bigarr[c[21],*]) else return
   if Nvector GT 22 then v23 = reform( bigarr[c[22],*]) else return
   if Nvector GT 23 then v24 = reform( bigarr[c[23],*]) else return
   if Nvector GT 24 then v25 = reform( bigarr[c[24],*]) else return
   if Nvector GT 25 then v26 = reform( bigarr[c[25],*]) else return
   if Nvector GT 26 then v27 = reform( bigarr[c[26],*]) else return
   if Nvector GT 27 then v28 = reform( bigarr[c[27],*]) else return
   if Nvector GT 28 then v29 = reform( bigarr[c[28],*]) else return
   if Nvector GT 29 then v30 = reform( bigarr[c[29],*]) else return
   if Nvector GT 30 then v31 = reform( bigarr[c[30],*]) else return
   if Nvector GT 31 then v32 = reform( bigarr[c[31],*]) else return
   if Nvector GT 32 then v33 = reform( bigarr[c[32],*]) else return
   if Nvector GT 33 then v34 = reform( bigarr[c[33],*]) else return
   if Nvector GT 34 then v35 = reform( bigarr[c[34],*]) else return
   if Nvector GT 35 then v36 = reform( bigarr[c[35],*]) else return
   if Nvector GT 36 then v37 = reform( bigarr[c[36],*]) else return
   if Nvector GT 37 then v38 = reform( bigarr[c[37],*]) else return
   if Nvector GT 38 then v39 = reform( bigarr[c[38],*]) else return
   if Nvector GT 39 then v40 = reform( bigarr[c[39],*]) else return
   if Nvector GT 40 then v41 = reform( bigarr[c[40],*]) else return
   if Nvector GT 41 then v42 = reform( bigarr[c[41],*]) else return
   if Nvector GT 42 then v43 = reform( bigarr[c[42],*]) else return
   if Nvector GT 43 then v44 = reform( bigarr[c[43],*]) else return
   if Nvector GT 44 then v45 = reform( bigarr[c[44],*]) else return
   if Nvector GT 45 then v46 = reform( bigarr[c[45],*]) else return
   if Nvector GT 46 then v47 = reform( bigarr[c[46],*]) else return
   if Nvector GT 47 then v48 = reform( bigarr[c[47],*]) else return
   if Nvector GT 48 then v49 = reform( bigarr[c[48],*]) else return
   if Nvector GT 49 then v50 = reform( bigarr[c[49],*]) else return
   if Nvector GT 50 then v51 = reform( bigarr[c[50],*]) else return
   if Nvector GT 51 then v52 = reform( bigarr[c[51],*]) else return
   if Nvector GT 52 then v53 = reform( bigarr[c[52],*]) else return
   if Nvector GT 53 then v54 = reform( bigarr[c[53],*]) else return
   if Nvector GT 54 then v55 = reform( bigarr[c[54],*]) else return
   if Nvector GT 55 then v56 = reform( bigarr[c[55],*]) else return
   if Nvector GT 56 then v57 = reform( bigarr[c[56],*]) else return
   if Nvector GT 57 then v58 = reform( bigarr[c[57],*]) else return
   if Nvector GT 58 then v59 = reform( bigarr[c[58],*]) else return
   if Nvector GT 59 then v60 = reform( bigarr[c[59],*]) else return
   if Nvector GT 60 then v61 = reform( bigarr[c[60],*]) else return
   if Nvector GT 61 then v62 = reform( bigarr[c[61],*]) else return
   if Nvector GT 62 then v63 = reform( bigarr[c[62],*]) else return
   if Nvector GT 63 then v64 = reform( bigarr[c[63],*]) else return
   if Nvector GT 64 then v65 = reform( bigarr[c[64],*]) else return
   if Nvector GT 65 then v66 = reform( bigarr[c[65],*]) else return
   if Nvector GT 66 then v67 = reform( bigarr[c[66],*]) else return
   if Nvector GT 67 then v68 = reform( bigarr[c[67],*]) else return
   if Nvector GT 68 then v69 = reform( bigarr[c[68],*]) else return
   if Nvector GT 69 then v70 = reform( bigarr[c[69],*]) else return
   if Nvector GT 70 then v71 = reform( bigarr[c[70],*]) else return
   if Nvector GT 71 then v72 = reform( bigarr[c[71],*]) else return
   if Nvector GT 72 then v73 = reform( bigarr[c[72],*]) else return
   if Nvector GT 73 then v74 = reform( bigarr[c[73],*]) else return
   if Nvector GT 74 then v75 = reform( bigarr[c[74],*]) else return
   if Nvector GT 75 then v76 = reform( bigarr[c[75],*]) else return
   if Nvector GT 76 then v77 = reform( bigarr[c[76],*]) else return
   if Nvector GT 77 then v78 = reform( bigarr[c[77],*]) else return
   if Nvector GT 78 then v79 = reform( bigarr[c[78],*]) else return
   if Nvector GT 79 then v80 = reform( bigarr[c[79],*]) 

  return
  end
